/*****************************************************************************
 *                                  _   _ ____  _     
 *  Project                     ___| | | |  _ \| |    
 *                             / __| | | | |_) | |    
 *                            | (__| |_| |  _ <| |___ 
 *                             \___|\___/|_| \_\_____|
 *
 * Copyright (C) 1998 Daniel Stenberg and Rafael Sagula
 * ------------------------------------------------------------
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *   MA 02111-1307, USA
 * ------------------------------------------------------------
 * Main author:
 * - Daniel Stenberg <Daniel.Stenberg@sth.frontec.se>
 *
 * 	http://www.fts.frontec.se/~dast/curl/
 *
 * $Source: /home/hardcore/dast/src/curl/RCS/main.c,v $
 * $Revision: 1.7 $
 * $Date: 1998/05/27 22:06:42 $
 * $Author: dast $
 * $State: Exp $
 * $Locker: dast $
 *
 * ------------------------------------------------------------
 * $Log: main.c,v $
 * Revision 1.7  1998/05/27 22:06:42  dast
 * corrected a spelling error
 *
 * Revision 1.6  1998/05/16 18:04:38  dast
 * allow user to specify -x proxy.server.com:1080 if desired
 *
 * Revision 1.5  1998/04/14 21:44:00  dast
 * New -P functionality
 * Uses CURL_ID
 *
 * Revision 1.4  1998/04/06 14:01:56  dast
 * corrected the spelling mistake "date" to "data"
 *
 * Revision 1.3  1998/03/23 15:22:33  dast
 * added -a awareness
 *
 * Revision 1.2  1998/03/23 10:36:27  dast
 * fixed header
 * added sagulas' changes
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "config.h"

#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

#include "url.h"

extern void hugehelp(void);

static void help(void)
{
  puts(CURL_ID "\n"
       "Usage: curl [options...] <url>\n"
       "Options: (H) means HTTP only (F) means FTP only\n"
       " -a/--append        Append to target file when uploading. (F)\n"
       " -A/--user-agent <string> User-Agent to send to server. (H)\n"
       " -c/--continue      Resume a previous transfer where we left it. (F)\n"
       " -C/--continue-at <offset> Specify absolute resume offset. (F)\n"
       " -d/--data          POST data. (H)\n"
       " -e/--referer       Referer page. (H)\n"
       " -f/--fail          Fail silently (no output at all) on errors. (H)\n"
       " -h/--help          Huge help text\n"
       " -i/--include       Include the HTTP-header in the output (H)\n"
       " -I/--head          Fetch the HTTP-header only (HEAD)! (H)\n"
       " -k/--keep-alive    Use Keep-Alive connection (H)\n"
       " -l/--list-only     List only names of an FTP directory (F)\n"
       " -m/--max-time <seconds> Maximum time allowed for the transfer\n"
       " -n/--netrc         Read .netrc for user name and password\n"
       " -o/--output <file> Write output to <file> instead of stdout\n"
       " -O/--remote-name   Write output to a file named as the remote file\n"
       " -p/--port <port>   Use port other than default for current protocol.\n"
       " -P/--ftpport <address> Use PORT with address instead of PASV when ftping. (F)\n"
       " -r/--range <range> Retrieve a byte range from a HTTP/1.1 server (H)\n"
       " -s/--silent        Silent mode. Don't show progress info\n"
       " -sl/--speed-limit  Stop transfer if below speed-limit for 'speed-time' secs\n"
       " -st/--speed-time   Time needed to trig speed-limit abort. Defaults to 30\n"
       " -t/--upload        Transfer/upload stdin to remote site. (F)\n"
       " -T/--upload-file <file> Transfer/upload <file> to remote site. (F)\n"
       " -u/--user <user:password> Specify user and password to use\n"
       " -U/--proxy-user <user:password> Specify Proxy authentication\n"
       " -v/--verbose       Makes the operation more talkative\n"
       " -V/--version       Outputs version number then quits\n"
       " -x/--proxy <host>  Use proxy. (Default port is 1080)"
       /* puts add a terminating newline by itself */
       );
}

struct LongShort {
  char *letter;
  char *lname;
};

static int argcheck(char option, int i, int argc)
{
  if(i >= argc -1 ) {
    printf("Option '%c' is missing a parameter!\n", option);
    help();
    return 1; 
  }
  else
    return 0;
}

int main(argc,argv)
    int argc;
    char *argv[];
{
  long conf=CONF_DEFAULT;
  unsigned short porttouse=0;

  char errorbuffer[URLGET_ERROR_SIZE];

  char *proxy = NULL;

  char *outfile = NULL;
  char *infile = NULL;
  char *url = NULL;
  char *useragent= "curl/" CURL_VERSION;
  char *userpwd=NULL;
  char *proxyuserpwd=NULL;
  char *range=NULL;
  char remotefile=FALSE;
  char *postfields=NULL;
  char *referer = NULL;
  char *ftpport = NULL;
  
  FILE *outfd = stdout;
  FILE *infd = stdin;
  char *urlbuffer=NULL;
  bool showerror=TRUE;
  long timeout=0;
  int infilesize=-1; /* -1 means unknown */

  int low_speed_limit=0;
  int low_speed_time=0;

  bool use_resume = FALSE;
  int resume_from = 0;

  int res;
  int i;

  struct LongShort aliases[]= {
    {"a", "append"},
    {"A", "user-agent"},
    {"c", "continue"},
    {"C", "continue-at"},
    {"d", "data"},
    {"e", "referer"},
    {"f", "fail"},
    {"h", "help"},
    {"i", "include"},
    {"I", "head"},
    {"k", "keep-alive"},
    {"l", "list-only"},
    {"m", "max-time"},
    {"n", "netrc"},
    {"o", "output"},
    {"O", "remote-name"},
    {"p", "port"},
    {"P", "ftpport"},
    {"r", "range"},
    {"s", "silent"},

    {"st", "speed-time"},
    {"sl", "speed-limit"},

    {"t", "upload"},
    {"T", "upload-file"},
    {"u", "user"},
    {"U", "proxy-user"},
    {"v", "verbose"},
    {"V", "version"},
    {"x", "proxy"}
  };
  if (argc < 2) {
    help();
    return URG_FAILED_INIT;
  }

  /* Parse options */
  for (i = 1; i < argc; i++) {
    if('-' == argv[i][0]) {
      char letter=0;
      char *nextl=NULL;
      if('-' == argv[i][1]) {
        /* try a long name */
        if(argv[i][2]) {
          int j;
          for(j=0; j< sizeof(aliases)/sizeof(aliases[0]); j++)
            if(strequal(aliases[j].lname, &argv[i][2])) {
              letter = *(aliases[j].letter);
	      nextl = &aliases[j].letter[1];
	      break;
            }
        }
      }
      else {
        letter = argv[i][1];
	nextl = &argv[i][2];
      }
      switch(letter) {
      case 'a':
	/* This makes the FTP sessions use APPE instead of STOR */
	conf |= CONF_FTPAPPEND;
	break;
      case 'A':
	/* This specifies the User-Agent name */
	if(argcheck(letter, i, argc)) /* check we have another argument */
	  return URG_FAILED_INIT;
	useragent = (char *)strdup (argv[++i]);
	break;
      case 'c':
	/* This makes us continue an ftp transfer */
	use_resume=TRUE;
	break;
      case 'C':
	/* This makes us continue an ftp transfer at given position */
	if(argcheck(letter, i, argc)) /* check we have another argument */
	  return URG_FAILED_INIT;
	resume_from= atoi(argv[++i]);
	use_resume=TRUE;
	break;
      case 'd':
        /* postfield data */
        postfields = argv[++i];
        /* printf("%s\n",postfields); fflush(stdout); */
        conf |= CONF_POST;
        break;
      case 'e':
        referer = argv[++i];
        conf |= CONF_REFERER;
        break;
      case 'f':
        /* fail hard on errors  */
        conf |= CONF_FAILONERROR;
        break;
      case 'h': /* h for HUGE help */
        hugehelp();
        return URG_FAILED_INIT;
      case 'i':
        conf |= CONF_HEADER; /* include the HTTP header as well */
        break;
      case 'I':
        conf |= CONF_HEADER; /* include the HTTP header in the output */
        conf |= CONF_NOBODY; /* don't fetch the body at all */
        break;
      case 'k':
        /* tell the server to keep the connection alive */
        conf |= CONF_KEEPALIVE;
        break;
      case 'l':
        conf |= CONF_FTPLISTONLY; /* only list the names of the FTP dir */
        break;
      case 'm':
        /* specified port */
        if(argcheck(letter, i, argc)) /* check we have another argument */
          return URG_FAILED_INIT;
        timeout = atoi(argv[++i]);
        break;
      case 'n':
	/* pick info from .netrc, if this is used for http, curl will
	   automatically enfore user+password with the request */
	conf |= CONF_NETRC;
	break;
      case 'o':
        /* output file */
        if(argcheck(letter, i, argc)) /* check we have another argument */
          return URG_FAILED_INIT;
        outfile = argv[++i]; /* write to this file */
        break;
      case 'O':
        /* output file */
        remotefile = TRUE;
        break;
      case 'P':
	/* This makes the FTP sessions use PORT instead of PASV */
	conf |= CONF_FTPPORT;
        if(argcheck(letter, i, argc)) /* check we have another argument */
          return URG_FAILED_INIT;
        /* use <eth0> or <192.168.10.10> style addresses. Anything except
           this will make us try to get the "default" address.
           NOTE: this is a changed behaviour since the released 4.1!
           */
        ftpport = argv[++i];
	break;
      case 'p':
        /* specified port */
        if(argcheck(letter, i, argc)) /* check we have another argument */
          return URG_FAILED_INIT;
        porttouse = atoi(argv[++i]);
        conf |= CONF_PORT; /* changed port */
        break;
      case 'r':
        /* byte range requested */
        if(argcheck(letter, i, argc)) /* check we have another argument */
          return URG_FAILED_INIT;
        range = argv[++i];
        conf |= CONF_RANGE;
        break;
      case 's':
	if(*nextl) {
	  if(argcheck(letter, i, argc)) /* check we have another argument */
	    return URG_FAILED_INIT;
	  
	  switch(*nextl) {
	  case 't':
	    low_speed_time = atoi(argv[++i]);
	    if(!low_speed_limit)
	      low_speed_limit = 1;
	    break;
	  case 'l':
	    low_speed_limit = atoi(argv[++i]);
	    if(!low_speed_time)
	      low_speed_time=30;
	    break;
	  default:
	    help(); /* show usage blurb */
	    return URG_FAILED_INIT;
	  }
	}
	else {
	  conf |= CONF_NOPROGRESS; /* don't show progress meter */
	  showerror=FALSE;
	}
        break;
      case 't':
        /* we are uploading */
        conf |= CONF_UPLOAD;
        break;
      case 'T':
        /* we are uploading */
        if(argcheck(letter, i, argc)) /* check we have another argument */
          return URG_FAILED_INIT;
        conf |= CONF_UPLOAD;
        infile = argv[++i]; /* read from this file */
        break;
      case 'u':
        /* user:password  */
        if(argcheck(letter, i, argc)) /* check we have another argument */
          return URG_FAILED_INIT;
        userpwd = argv[++i];
        conf |= CONF_USERPWD;
        break;
      case 'U':
        /* Proxy user:password  */
        if(argcheck(letter, i, argc)) /* check we have another argument */
          return URG_FAILED_INIT;
        proxyuserpwd = argv[++i];
        conf |= CONF_PROXYUSERPWD;
        break;
      case 'v':
        conf |= CONF_VERBOSE; /* talk a lot */
        break;
      case 'V':
        puts (CURL_ID);
        return URG_FAILED_INIT;
      case 'x':
	{
	  char * prox_portno;

	  /* proxy */
	  if(argcheck(letter, i, argc)) /* check we have another argument */
	    return URG_FAILED_INIT;
	  proxy = argv[++i];
	  conf |= CONF_PROXY;

	  /* allow user to specify -x proxy.server.com:1080 if desired */
	  prox_portno = strchr (proxy, ':');
	  if (prox_portno) {
	    *prox_portno = 0x0;
	    prox_portno ++;
	    porttouse = atoi(prox_portno);
	    conf |= CONF_PORT;
	  }
	  /* if the portno is NOT specified, then do NOT try to guess that
	   * the default will be 1080.  In fact, picking port numbers correctly
	   * is somewhat tricky if SSL & proxy are used together. 
	   */
	}
      break;

      default: /* unknown flag */
        if(letter)
          printf("Unknown option '%c'\n", letter);
        help(); /* show usage blurb */
        return URG_FAILED_INIT;
      }
    }
    else {
      url = argv[i];
      break;
    }
  }


  if(!url) {
    fprintf(stderr, "%s: consider specifying an URL too! :-)\n", argv[0]);
    return URG_FAILED_INIT;
  }

  if(outfile && infile) {
    fprintf(stderr, "%s: you can't both upload and download!\n", argv[0]);
    return URG_FAILED_INIT;
  }
 
  if (outfile || remotefile) {
    /* 
     * We have specified a file name to store the result in, or we have
     * decided we want to use the remote file name.
     */

    if(remotefile) {
      /* Find and get the remote file name */
      outfile=strstr(url, "://");
      if(outfile)
        outfile+=3;
      else
        outfile=url;
      outfile = strrchr(outfile, '/');
      if(!outfile || !strlen(++outfile)) {
        fprintf(stderr, "%s: Remote file name has no length!\n", argv[0]);
        return URG_WRITE_ERROR;
      }
    }

    if((0 == resume_from) && use_resume) {
      /* we're told to continue where we are now, then we get the size of the
	 file as it is now and open it for append instead */
      struct stat fileinfo;

      if(0 == stat(outfile, &fileinfo)) {
	/* set offset to current file size: */
	resume_from = fileinfo.st_size;
      }
      /* else let offset remain 0 */
    }

    /* open file for output: */
    outfd=(FILE *) fopen(outfile, resume_from?"a":"w");
    if (!outfd) {
      fprintf(stderr, "%s: Can't open '%s'!\n", argv[0], outfile);
      return URG_WRITE_ERROR;
    }
  }
  if (infile) {
    /*
     * We have specified a file to upload
     */
    struct stat fileinfo;

    /* If no file name part is given in the URL, we add this file name */
    char *ptr=strstr(url, "://");
    if(ptr)
      ptr+=3;
    else
      ptr=url;
    ptr = strrchr(ptr, '/');
    if(!ptr || !strlen(++ptr)) {
      /* The URL has no file name part, add the local file name. In order
         to be able to do so, we have to create a new URL in another buffer.*/
      urlbuffer=malloc(strlen(url) + strlen(infile) + 3);
      if(!urlbuffer) {
        fprintf(stderr, "%s: out of memory\n", argv[0]);
        return URG_OUT_OF_MEMORY;
      }
      if(ptr)
        /* there is a trailing slash on the URL */
        sprintf(urlbuffer, "%s%s", url, infile);
      else
        /* thers is no trailing slash on the URL */
        sprintf(urlbuffer, "%s/%s", url, infile);

      url = urlbuffer; /* use our new URL instead! */
    }

    infd=(FILE *) fopen(infile, "r");
    if (!infd) {
      fprintf(stderr, "%s: Can't open '%s'!\n", argv[0], infile);
      return URG_READ_ERROR;
    }
    else if(stat(infile, &fileinfo)) {
      fprintf(stderr, "%s: Can't stat '%s'!\n", argv[0], infile);
      return URG_READ_ERROR;
    }
    infilesize=fileinfo.st_size;

  }
  if((conf&CONF_UPLOAD) && use_resume && (0==resume_from)) {
    resume_from = -1; /* -1 will then force get-it-yourself */
  }

  /* This was previously done in urlget, but that was wrong place to do it */
  if(isatty(fileno(outfd)))
    /* we send the output to a tty, and therefor we switch off the progress
       meter right away */
    conf |= CONF_NOPROGRESS;

  res = urlget(URGTAG_FILE, outfd,  /* where to store */
               URGTAG_INFILE, infd, /* for uploads */
               URGTAG_INFILESIZE, infilesize, /* size of uploaded file */
               URGTAG_URL, url,     /* what to fetch */
               URGTAG_PORT, porttouse, /* from which port */
               URGTAG_PROXY, proxy, /* proxy to use */
               URGTAG_FLAGS, conf, /* flags */
               URGTAG_USERPWD, userpwd, /* user + passwd */
               URGTAG_PROXYUSERPWD, proxyuserpwd, /* Proxy user + passwd */
               URGTAG_RANGE, range, /* range of document */
               URGTAG_ERRORBUFFER, errorbuffer,
               URGTAG_TIMEOUT, timeout,
               URGTAG_POSTFIELDS, postfields,
               URGTAG_REFERER, referer,
               URGTAG_USERAGENT, useragent,
               URGTAG_FTPPORT, ftpport,
	       URGTAG_LOW_SPEED_LIMIT, low_speed_limit,
	       URGTAG_LOW_SPEED_TIME, low_speed_time,
	       URGTAG_RESUME_FROM, use_resume?resume_from:0,
               URGTAG_DONE); /* always terminate the list of tags */

  if((res!=URG_OK) && showerror)
    fprintf(stderr, "%s: %s\n", argv[0], errorbuffer);

  if(urlbuffer)
    free(urlbuffer);
  if (outfile)
    fclose(outfd);
  if (infile)
    fclose(infd);

  return(res);
}

